library(arm)
library(digest)
library(rdict)
	
LD.MeasuresHashed <- function (donnees, doBlock=1, totalBlocks=1, cachePrevOnly=FALSE, dumpFile=NULL) 
{
	arrLociNames <- colnames(donnees);
	nLoci <- ncol(donnees);
	
	nComparisons <- 0;
	arrRetCache <- rdict_new();
	
	nTotalCells <- (nLoci*nLoci- nLoci) / 2;
	nCellsPerBlock <- floor(nTotalCells/totalBlocks);
	nLeftCells <- nTotalCells %% nCellsPerBlock;
	
	nFrom <- (doBlock - 1) * nCellsPerBlock + 1;
	nTo <- nFrom + nCellsPerBlock -1;
	if ( doBlock == totalBlocks) {
		nTo <- nTo + nLeftCells;
	}
	
	nDone <- 0;
	nPrevDoneProgress <- 0;
	nTotalNeedtoDo <- nTo - nFrom + 1;
	
	#write header:
	
		
	if (!is.null(dumpFile)) {
		if (doBlock != 1) {
			write.table("", dumpFile, append=FALSE, col.names=FALSE, quote=FALSE, sep="\t", row.names=FALSE);
		} else {
		
			arrHeader <- t(as.matrix(c("loc1", "loc2", "t", "p", "approximated", "included_ind")));
			
			write.table(arrHeader, dumpFile, append=FALSE, col.names=FALSE, quote=FALSE, sep="\t", row.names=FALSE);
		}
		
	}
	
	
	for (i in seq(1, nLoci-1)) {
		for (j in seq(i+1, nLoci)) {
		
			nComparisons <-  nComparisons+1;
			
			if ( nComparisons < nFrom || nComparisons > nTo) {
				next; #skip this.
			}
			
			nDone <- nDone + 100;
			
			nCurrProgress <- floor(nDone  / nTotalNeedtoDo);
			
			if (nCurrProgress %% 1 ==0) {
				if (nCurrProgress != nPrevDoneProgress) {
					cat(paste("B", doBlock, " ", nCurrProgress,"% \n", sep=""));
					nPrevDoneProgress <- nCurrProgress;
					
					# clear cache periodically to speed up look-ups
					arrRetCache <- NULL;
					arrRetCache <- rdict_new();
				}
			}
			
			arrLociPair <- cbind(donnees[,i] , donnees[,j]);
			arrLociPairNoN <- arrLociPair[ complete.cases(arrLociPair) , ]; #remove individuals with at least one NA
			nIncludedInd <- nrow(arrLociPairNoN);
			
			#hash the patterns
			chksum_i <-  digest(as.vector(arrLociPairNoN[,1] ), algo="sha1");# paste(li, sep='', collapse='');
			chksum_j <- digest(as.vector(arrLociPairNoN[,2]), algo="sha1"); #paste(lj, sep='', collapse='');
			
			sortedChkSums <- sort(c(chksum_i, chksum_j));
			sKey <- paste(sortedChkSums, sep="", collapse='');
			
			arrRetCache_lv2 <- rdict_get( arrRetCache, sortedChkSums[1]); #is the first sequence in dict?
			if (is.null(arrRetCache_lv2)) {
				arrRetCache_lv2 <- rdict_new();
				rdict_put(arrRetCache, sortedChkSums[1], arrRetCache_lv2); #create second level dictionary
					 
			}
				
			cachedRet <- rdict_get(arrRetCache_lv2, sortedChkSums[2]);
			bKeyNotFound <- (is.null(cachedRet));
				
			if (bKeyNotFound) {
			
				cachedRet <- (MeasureLogit(arrLociPairNoN[,1] ,arrLociPairNoN[,2] ));
				rdict_put( rdict_get(arrRetCache, sortedChkSums[1] ) , sortedChkSums[2], cachedRet);
				
			} else {
				
			}
			
			#Dump To File
			write.table(paste(arrLociNames[i],arrLociNames[j],cachedRet[["t"]],cachedRet[["p"]],cachedRet[["approx"]],cachedRet[["ind"]], sep="\t" ), dumpFile, append=TRUE, col.names=FALSE, quote=FALSE, sep="", row.names=FALSE);
			
		}
	}
}

MeasureLogit <- function(val1 , val2) {
	ret <- list();
	nN <- length(val1);
	#cat("Included individuals:", nN);
	ret[["ind"]] <- nN;
	
	ret <- tryCatch(
	{
		model <- bayespolr(as.factor(val1) ~ val2)
		nT=coef(summary(model))[7]
		p_pop1=2*pt(-abs(nT),df=nN-2)
		ret[["t"]] <- nT; 
		ret[["p"]] <- p_pop1;
		ret[["approx"]] <- "n";
		ret;
	} , 
	warning=function(w) {
		cat(w);
		ret[["t"]] <- NA; 
		ret[["p"]] <- NA;
		ret[["approx"]] <- NA;
		ret;
	} , 
	error = function(e) {
		#check stuff:
		nVar1Lack <- LackWhichGenotype(val1);
		nVar2Lack <- LackWhichGenotype(val2);
		nFake1 <- 0;
		nFake2 <- 0;
		
		#print(nVar1Lack);
		#print(nVar2Lack);
		if (length(nVar1Lack) != 0 && length(nVar2Lack) != 0) {
			ret[["approx"]] <- "y"; # y means added fake point
			if (length(nVar1Lack) == 0) {
				nFake1 <-1;
			} else {
				nFake1 <- nVar1Lack[[1]];
			}
			
			
			if (length(nVar2Lack) == 0) {
				nFake2 <-1;
			} else {
				nFake2 <- nVar2Lack[[1]];
			}
			
			if (nFake1 == nFake2) {
				if (nFake2 == 1) {
					nFake1 = 0;
				} else if (nFake2==0) {
					nFake1 = 2;
				}
				else {
					nFake1 = 0;
				}
			}
			
			val1<-c(nFake1,val1)
			val2<-c(nFake2,val2)
			#print(val1);
			#print(val2);
		}	else if (length(nVar1Lack) != 0 && length(nVar2Lack) == 0) { # we can switch the two columns.
			valtmp <- val1;
			val1 <- val2;
			val2 <- valtmp;
			ret[["approx"]] <- "s"; #s means switch col.
		} else { #doesn't matter.
			
		}
		
		
		ret <- tryCatch(
			{
				model <- bayespolr(as.factor(val1) ~ val2)
				nT=coef(summary(model))[7]
				p_pop1=2*pt(-abs(nT),df=nN-2)
				ret[["t"]] <- nT; 
				ret[["p"]] <- p_pop1;
				
				ret;
			}
			,
			warning=function(w) {
				#print(w);
				ret[["t"]] <- NA; 
				ret[["p"]] <- NA;
				ret[["approx"]] <- NA;
				ret;
			} 	
			,
			error=function(e) {
				#print(e);
				ret[["t"]] <- NA; 
				ret[["p"]] <- NA;
				ret[["approx"]] <- NA;
				ret;
			}
			,
			finally=function() {}
		)
		
		ret;
	}, 
	finally= {

		}

	)
	
	return(ret);
}

vPossibleGenotypes <- c(0, 1, 2);
LackWhichGenotype <- function(vGenotype) {
	return(setdiff( vPossibleGenotypes, unique(as.numeric(vGenotype))  ));
}
